/*--------------------------------------------------------------------.
| parse.c							      |
| 								      |
| Copyright (c) 1997 Jonathan Hudson				      |
| 								      |
| You may distribute under the terms of either the GNU General Public |
| License or the Artistic License, as specified in the README file.   |
`--------------------------------------------------------------------*/
static char rcsid[] = "$Id$";
/*
 * $Log$: 
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <qdos.h>

#include "qfm.h"
#define NOT_FOUND (-7)
#define QFAX_PHONE "qfax_phone"
#define OLD_BOOK
#define QFAX_FILE "qfax_dat"

#ifndef offsetof
#define offsetof(sname,fname) ((long)&((sname *)0)->fname)
#endif

typedef struct
{
    char *id;
    void (*func)(void *, void *, int);
    long offset;
    int  mval;
    short flag;
} PARSELIST;

typedef struct _alias
{
    char   phone[40];
    char   alias[40];
    char   abbrev[16];
    char   dcode[40];
    unsigned short  flag;
} ALIAS;

static FAXDATA defs = {
    NULL,
    "ram1_fax_",
    "ram2_fax_log",
    "ram2_",
    "_fax",
    "qfax_phone",
    {0},
    {0}
};

#ifdef DEBUG
Report (char *ctl,...)
{
    va_list va;
    char t[128];
    va_start (va, ctl);
    vsprintf (t, ctl, va);
    va_end (va);
    io_sstrg (0x00010001, -1, t, (short) strlen (t));
}

void *xalloc (int i)
{
    void *p;
    if ((p = malloc (i)))
    {
	memset (p, '\0', i);
    }
    else
    {
	exit (-3);
    }
    return p;
}

void xfree (void *q)
{
    if (q)
	free (q);
}

char *strdup (char *p)
{
    char *q;
    q = xalloc (strlen (p) + 1);
    strcpy (q, p);
    return q;
}
#endif


static char *strim (char *s)
{
    char *p;
    short n;

    n = strlen(s);
    p = s + n;
    while(*--p <= ' ' && n--)
    {
        *p = '\0';
    }
    return s;
}

static short AlMatch(char *s, char **p2)
{
    static char *opts[] = {"TSI","NAME", "ALIAS","DCODE", "FULL", "2D", NULL};
    char **p;
    short n = 0;
    short sl;
    for(p = opts; *p; p++)
    {
        *p2 = NULL;
        if(strnicmp(s, *p, (sl = strlen(*p))) == 0)
        {
            {
                (*p2) = s + sl + strspn(s + sl, "= \t");
                break;
            }
        }
        n++;
    }
    return n;
}

ins_list (LIST ** pHead, LIST *pi, int (*func) ())
{
    LIST *pl = NULL;
    LIST *pz;

    if (*pHead == NULL)
    {
        *pHead = pi;
        pi->next = NULL;
    }
    else
    {
        pl = NULL;
        for (pz = *pHead; pz; pz = pz->next)
        {
            if (func)
            {
                int n;

                n = (*func) (pi->item, pz->item);
                if (n < 0)
                {
                    if (pl)
                    {
                        pl->next = pi;
                    }
                    else
                    {
                        *pHead = pi;
                    }
                    pi->next = pz;
                    return 0;
                }
            }
            pl = pz;
        }
        pl->next = pi;
        pi->next = NULL;
    }
}

static void  pAlias (FAXDATA *p, char *s, int nx)
{
    char *p0, *p1,*p2;
    short n,ok;
    ALIAS pn;

    pn.flag = 0;
    ok = 1;
    for(p0 = s; (p1 = strtok(p0, ",")); p0 = NULL)
    {
	p1 += strspn (p1, " \t");
	n = AlMatch(p1, &p2);                            
	switch(n)
	{
	    case 0:
		strncpy (pn.phone, p2, 39);
		break;
	    case 1:
		strncpy (pn.alias, p2, 39);
		break;
	    case 2:
		strncpy (pn.abbrev, p2, 15);
		break;                
	    case 3:
		strncpy (pn.dcode, p2, 39);
		pn.flag &= (short) ~0x8000;
		break;
	    case 4:
		strncpy (pn.dcode, p2, 39);
		pn.flag |= 0x8000;
		break;
	    case 5:
		pn.flag |= 1;
		break;
	    default:
		ok = 0;
		break;
	}
    }
    if(ok)
    {
	LIST *l;
	char csi[40],txt[80];
	int nl;

	if (pn.flag & 0x8000)
	{
	    strcpy (csi, pn.dcode);
	}
	else
	{
	    if (*pn.dcode)
	    {
		if (*pn.dcode == '-')
		{
		    strcpy (csi, pn.dcode + 1);
		}
		strcpy (csi, pn.dcode);
	    }
	    else
	    {
		*csi = '\0';
	    }
	    strcat (csi, pn.phone);
	}

	{
	    char *u;
	    u = stpcpy(txt, pn.alias);
	    while(u < txt+33)
	    {
		*u++ = ' ';
	    }
	    u = stpcpy(u, csi);
	    nl = u - txt;
	    l = xalloc(sizeof(LIST)+nl+1);
	    strcpy(l->item, txt);
	    ins_list((LIST **)&p->list, l, strncmp);
	}
    }
}

static void  pPhBook (FAXDATA *p, char *s, int nx)
{
    int fd;
    char buf[256];

    strcpy(p->phonebook, s);
    if((fd = io_open(s, OLD_EXCL)) > 0)
    {
        while(io_fline (fd, -1, buf, sizeof(buf)) > 0)
        {
            char *pe;
            if((pe = strchr(buf, '\n')))
            {
                *pe = 0;
            }
	    if(*buf)
	    {
		pAlias(p, buf, 0);
	    }
        }
        io_close(fd);
    }
}

static void  pString (char *p, char *s, int n)
{
    if(*s)
    {
        strncpy(p, s, n);
    }
}

static void  pNewStr (char *p, char *s, int n)
{
    if(*s)
    {
	*(char **)p = strdup(s);
    }
}

# define VF (void(*)(void *,void *,int))

static PARSELIST pl[] = {
    {"IND",     VF pString,    offsetof(FAXDATA, in_dir),      39},
    {"INF",     VF pString,    offsetof(FAXDATA, info_file),   39},
    {"SUFFIX",  VF pString,    offsetof(FAXDATA, suffix),      15},
    {"PHONEBOOK", VF pPhBook},

    {"LMB_EDIT",    VF pNewStr, offsetof(FAXDATA, lmb[0])}, 
    {"LMB_MAKE",    VF pNewStr, offsetof(FAXDATA, lmb[1])}, 
    {"LMB_VIEW",    VF pNewStr, offsetof(FAXDATA, lmb[2])}, 
    {"LMB_SEND",    VF pNewStr, offsetof(FAXDATA, lmb[3])}, 
    {"LMB_PRINT",   VF pNewStr, offsetof(FAXDATA, lmb[4])}, 
    {"RMB_RECEIVE", VF pNewStr, offsetof(FAXDATA, rmb[5])}, 

    {"LMB_EDDAT",   VF pNewStr, offsetof(FAXDATA, lmb[7])}, 
    {"LMB_EDPHONE", VF pNewStr, offsetof(FAXDATA, lmb[8])}, 
    {"LMB_VLOG",    VF pNewStr, offsetof(FAXDATA, lmb[9])}, 
    {"LMB_VINFO",   VF pNewStr, offsetof(FAXDATA, lmb[10])}, 

    {"RMB_EDIT",    VF pNewStr, offsetof(FAXDATA, rmb[0])}, 
    {"RMB_MAKE",    VF pNewStr, offsetof(FAXDATA, rmb[1])}, 
    {"RMB_VIEW",    VF pNewStr, offsetof(FAXDATA, rmb[2])}, 
    {"RMB_SEND",    VF pNewStr, offsetof(FAXDATA, rmb[3])}, 
    {"RMB_PRINT",   VF pNewStr, offsetof(FAXDATA, rmb[4])}, 
    {"RMB_RECEIVE", VF pNewStr, offsetof(FAXDATA, rmb[5])}, 

    {"RMB_EDDAT",   VF pNewStr, offsetof(FAXDATA, rmb[7])}, 
    {"RMB_EDPHONE", VF pNewStr, offsetof(FAXDATA, rmb[8])}, 
    {"RMB_VLOG",    VF pNewStr, offsetof(FAXDATA, rmb[9])}, 
    {"RMB_VINFO",   VF pNewStr, offsetof(FAXDATA, rmb[10])}, 
   {NULL, NULL}
};

void FaxParams (FAXDATA **p)
{
    int fp;
    char *pf, *pfe;
    static const char *sdefs[] = 
    {
	"qed %f$QFAX_INDIR_txt",
	"text2g3 %f",
	"qfv -l",
	"qfax -v -s %t %f$QFAX_TEMP$QFAX_SUFFIX",
	"printfax %f", 
	"qfax -Rv",
	"*",
	"qed $QFAX_DATA",
	"qed $QFAX_PHONE",
	"qless %?",
	"qless $QFAX_INFO",
	NULL
    };

    if(*p == NULL)
    {
	*p = xalloc(sizeof(FAXDATA));

	if((pf = pfe = getenv("QFAX_DATA")) == NULL)
	{
	    pf = QFAX_FILE;
	}
	**p = defs;
       
	if((fp = io_open(pf, OLD_EXCL)) > 0)
	{
	    char text [128];
	    int rv;
	    
	    while((rv = io_fline(fp, -1, text, sizeof(text))) > 0)
	    {
		char *s;
		PARSELIST *ppl;
		char *ptr;
		
		if((s = strchr(text,'\n'))) *s = 0;
		ptr = *(char **)p;
		strim(text);
		for(ppl = pl; ppl->id; ppl++)
		{
		    if(strnicmp(text, ppl->id, strlen(ppl->id)) == 0)
		    {
			int l = 0;
			if((s = strchr(text, '=')))
			{
			    s++;
			    l = strspn(s, " \t");
			}
			else if(ppl->flag)
			{
			    s = text+strlen(text);
			    l = 0;
			}
			if(s)
			{
			    (ppl->func)((ptr + ppl->offset), s + l, ppl->mval);
			    break;
			}
		    }
		}
	    }
	    io_close(fp);
	}
	{
	    char **s,**l,**r;
	    for(s = sdefs, l = (*p)->lmb, r = (*p)->rmb;
		*s; s++, l++, r++)
	    {
		if(*l  == NULL)
		{
		    *l  = strdup(*s);
		}
		if(*r == NULL)
		{
		    *r = *l;
		}
	    }
	}
	xfree(pfe);
    }
}

void FreeParams(FAXDATA **p)
{
    void *dead;
    char **l,**r;

    for(l = (*p)->lmb, r = (*p)->rmb; *l; l++, r++)
    {
	if(*r != *l) xfree(*r);
	xfree(*l);
    }

    while((*p)->list)
    {
	dead = (*p)->list;
	(*p)->list = (*p)->list->next;
	xfree(dead);
    }
    xfree(*p);
    *p = NULL;
}

ListBook(char *p, FAXDATA *ff)
{
    int fd;
    if(p)
    {
	fd = io_open(p, NEW_OVER);
    }
    else
    {
	fd = 0x00010001;
    }
    if(ff->list)
    {
        LIST *pl;
        for(pl = ff->list; pl; pl = pl->next)
        {
	    io_sstrg(fd, -1, pl->item, strlen(pl->item));
	    io_sbyte(fd, -1, '\n');
	}
    }
    if(p)
	io_close(fd);
}

#ifdef DEBUG
int main(int ac, char **av)
{
    FAXDATA *ff;
    char **p;
    short n;

    FaxParams(&ff);
    Report ("In bound        : %s\n", ff->in_dir);
    Report ("Info File       : %s\n", ff->info_file);
    Report ("Temp            : %s\n", ff->Temp);
    Report ("Book            : %s\n", ff->phonebook);
    Report ("List            : %s\n", ff->list);

    if(ff->list)
    {
        LIST *pl;
	ListBook(NULL, ff);
    }

    for(n = 0, p = ff->lmb; *p; p++, n++)
    {
	Report ("Lmb[%d]       : %s\n", n, *p);
    }

    for(n = 0, p = ff->rmb; *p; p++, n++)
    {
	Report ("Rmb[%d]       : %s\n", n, *p);
    }

    return 0;
}
int (*_Cstart) () = main;
#endif
